/*******
svmvia, an implementation of the full regularization path for support vector machines
Copyright (C) 2007 Oliver Serang
Written by Oliver Serang at the Department of Genome Sciences at the University of Washington

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*********/

#include "matrix.h"

using namespace std;

#include <iostream>

// matrix inverter

ostream & operator << (ostream & os, matrix m)
{
  for (int k=0; k < m.r; k++)
    {
      os << m[k] << endl;
    }
  return os;
}

	matrix::matrix()
	{
		r=0;
		c=0;
	}
	
	matrix::matrix(int rr, int cc):
	data(rr)
	{
		for (int rg=0; rg<rr; rg++)
		{
			data[rg]=seq(cc);
		}
		r=rr;
		c=cc;
	}

	seq & matrix::operator [](int k)
	{
		return data[k];
	}

seq matrix::operator [] (int k) const 
{
  return data[k];
}

	const matrix & matrix::operator *=(matrix & rhs)
	{
		if (c!=rhs.r)
		{
		  cerr << "dimension mismatch for matrix *=" << endl;
		  throw;
		}
		matrix temp(r,rhs.c);
		for (int k=0; k<r; k++)
		{
			for (int j=0; j<rhs.c; j++)
			{
				temp[k][j]=0;
				for (int i=0; i<c; i++)
				{
					temp[k][j]+=data[k][i]*rhs[i][j];
				}
			}
		}
		*this = temp;
		return *this;
	}

	const matrix & matrix::operator +=(matrix & rhs)
	{
	  if (r != rhs.r && c != rhs.c)
	    {
	      cerr << "dimension mismatch for matrix +=" << endl;
	      throw;
	    }
		for (int k=0; k<r; k++)
		{
		  data[k] += rhs[k];
		}
		return *this;
	}

	matrix matrix::operator +(matrix & rhs)
	{
		matrix temp=*this;
		return temp+=rhs;
	}

	matrix matrix::operator *(matrix & rhs)
	{
		matrix temp=*this;
		return temp*=rhs;
	}

	matrix matrix::operator *(double rhs)
	{
		matrix temp=*this;
		return temp*=rhs;
	}

matrix matrix::operator [](const set & rhs)
{
  int n = rhs.size();
  matrix temp(n,n);
  int k,j;
  for (k=0; k<n; k++)
    {
      for (j=0; j<n; j++)
	{
	  temp[k][j] = data[ rhs[k] ][ rhs[j] ];
	}
    }
  return temp;
}

void matrix::resize(int rr, int cc)
{
  data.resize(rr);
  r = rr;
  c = cc;
  for (int k=0; k<rr; k++)
    {
      data[k].resize(cc);
    }
}

matrix matrix::restrictCol(const set & rhs)
{
  matrix ret(r, rhs.size());
  int k,j;
  for (k=0; k<r; k++)
    {
      for (j=0; j<rhs.size(); j++)
	{
	  ret[k][j] = data[k][ rhs[j] ];
	}
    }
  return ret;
}

void matrix::removeRC(int i)
{
  // only works for a square matrix

  //cout << "RC REMOVAL: " << i << endl;
  //cout << *this << endl;

  if (i == r-1 && i == c-1)
    {
      // removing last- don't bother shifting anything
      resize(r-1, c-1);
      return;
    }

  int k,j;
  for (k=i; k<(r-1); k++)
    {
		for (j=i; j<(c-1); j++)
		{
		  data[k][j] = data[k+1][j+1];
		}
    }

	for (k=i; k<(r-1); k++)
	{
	  for (j=0; j<i; j++)
		{
		  data[k][j] = data[k+1][j];
		}
	}

  for (k=0; k<i; k++)
    {
		for (j=i; j<(c-1); j++)
		{
			data[k][j] = data[k][j+1];
		}
    }
  resize(r-1, c-1);
  
  //  cout << "POST RC" << endl;
  //  cout << *this << endl << endl;
}

seq matrix::operator *(const seq & rhs)
{
  // this uses a pretty needless copy
  // in order to reuse code.
  // speed up can be had.

  // rXc * cX1 -> rX1
  matrix temp(rhs.size(), 1);
  int k;
  for (k=0; k<rhs.size(); k++)
    {
      temp[k][0] = rhs[k];
    }

  matrix resMat = (*this) * temp;
  seq res(r);
  for (k=0; k<r; k++)
    {
      res[k] = resMat[k][0];
    }
  return res;
}

	const matrix & matrix::operator *=(double rhs)
	{
		for (int k=0; k<r; k++)
		{
		  (*this)[k] *= rhs;
		}
		return *this;
	}

	matrix matrix::operator -()
	{
		matrix temp(r,c);
		for (int k=0; k<r; k++)
		{
		  temp[k] = - (*this)[k];
		}
		return temp;
	}

	void matrix::invert()
	{
	  //	  cout << "Inv: " << r << endl;
		if (r!=c)
		{
			// error, non-square
		}
		matrix temp=id(r);
		int k;

		// this is just a quick heuristic to swap rows
		// if a given row has a small cell in the diagonal
		/***
		for (k=0; k<r; k++)
		{
			if (fabs(data[k][k]) < 1e-8)
			{
				for (int j=k+1; j<r; j++)
				{
					if (data[j][k]!=0)
					{
						swap(temp[j],temp[k]);
						swap(data[j],data[k]);
						break;
					}
				}
			}
		}
		***/
		for (k=0; k<r; k++)
		  {
		    data[k][k] += 1e-6;
		  }

		//     		cout << "MATRIX POST SWAPPING:" << endl;
		//		cout << *this << endl << endl;
		//		cout << temp << endl << endl;
		//		cout << "<END POST MATRIX SWAP PRINT>" << endl;

		for (k=0; k<r; k++)
		{
			temp[k] /= data[k][k];
			data[k] /= data[k][k];
			for (int j=0; j<r; j++)
			{
				if (j==k) continue;

				seq row = data[k];
				seq idRow = temp[k];

				idRow *= -data[j][k];
				temp[j] += idRow;

				row *= -data[j][k];
				data[j] += row;
			}
		}
		for (k=0; k<r; k++)
		{
			for (int j=0; j<c; j++)
				data[k][j]=temp[k][j];
		}
	}

	matrix matrix::id(int R)
	{
		matrix temp(R,R);
		for (int k=0; k<R; k++)
		{
			for (int j=0; j<R; j++)
			{
				if (j!=k)
					temp[k][j]=0;
				else
					temp[k][j]=1;
			}
		}
		return temp;
	}
