/*******
svmvia, an implementation of the full regularization path for support vector machines
Copyright (C) 2007 Oliver Serang
Written by Oliver Serang at the Department of Genome Sciences at the University of Washington

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*********/

#include "seq.h"

double & seq::operator [](int k)
{
  return data[k];
}

seq seq::operator [](const set & rhs)
{
  seq ret(rhs.size());
  int k;
  for (k=0; k<ret.size(); k++)
    {
      // round off the double to get the closest integer index
      ret[k] = (*this)[ int(rhs[k] +.5) ];
    }
  return ret;
}

int seq::size() const
{
  return data.size();
}

void seq::error(char * s)
{
  cerr << s << endl;
  throw;
}

void seq::check_size(const seq & rhs)
{
  if (size() != rhs.size())
    {
      error("mismatch in sequence sizes");
    }
}

const seq & seq::operator +=(double rhs)
{
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] += rhs;
    }
  return *this;
}

const seq & seq::operator -=(double rhs)
{
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] -= rhs;
    }
  return *this;
}

const seq & seq::operator *=(double rhs)
{
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] *= rhs;
    }
  return *this;
}

const seq & seq::operator /=(double rhs)
{
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] /= rhs;
    }
  return *this;
}

seq seq::operator +(double rhs) const
{
  seq temp = *this;
  temp += rhs;
  return temp;
}

seq seq::operator -(double rhs) const
{
  seq temp = *this;
  temp -= rhs;
  return temp;
}

seq seq::operator *(double rhs) const
{
  seq temp = *this;
  temp *= rhs;
  return temp;
}

seq seq::operator /(double rhs) const
{
  seq temp = *this;
  temp /= rhs;
  return temp;
}

seq seq::operator -() const
{
  int k;
  seq temp = *this;
  for (k=0; k<size(); k++)
    {
      temp[k] = -temp[k];
    }
  return temp;
}

const seq & seq::operator +=(const seq & rhs)
{
  check_size(rhs);
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] += rhs[k];
    }
  return *this;
}

const seq & seq::operator -=(const seq & rhs)
{
  check_size(rhs);
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] -= rhs[k];
    }
  return *this;
}

const seq & seq::operator *=(const seq & rhs)
{
  check_size(rhs);
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] *= rhs[k];
    }
  return *this;
}

const seq & seq::operator /=(const seq & rhs)
{
  check_size(rhs);
  int k;
  for (k=0; k<size(); k++)
    {
      data[k] /= rhs[k];
    }
  return *this;
}

seq seq::operator +(const seq & rhs) const
{
  seq temp = *this;
  temp += rhs;
  return temp;
}

seq seq::operator -(const seq & rhs) const
{
  seq temp = *this;
  temp -= rhs;
  return temp;
}

seq seq::operator *(const seq & rhs) const
{
  seq temp = *this;
  temp *= rhs;
  return temp;
}

seq seq::operator /(const seq & rhs) const
{
  seq temp = *this;
  temp /= rhs;
  return temp;
}

int seq::numMatch(double val) const
{
  int k, count = 0;
  for (k=0; k<size(); k++)
    {
      if ((*this)[k] == val)
	count++;
    }
  return count;
}

set seq::match(double val) const
{
  // you may want to rewrite this to compare over epsilon

  int k;
  int count = numMatch(val);
  set ret(count);

  int c = 0;
  for (k=0; k<size(); k++)
    {
      if ( (*this)[k] == val)
	{
	  ret[c] = k;
	  c++;
	}
    }
  return ret;
}

set seq::maxSet()
{
  return match( max() );
}

set seq::minSet()
{
  return match( min() );
}

double seq::max()
{
  double big = -1e20;
  for (int k=0; k<size(); k++)
  {
    if (data[k] > big)
      big = data[k];
  }
  return big;
}

double seq::min()
{
  double small = 1e20;
  for (int k=0; k<size(); k++)
  {
    if (data[k] < small)
      small = data[k];
  }
  return small;
}

void seq::setAll(double d)
{
  for (int k=0; k<size(); k++)
    data[k] = d;
}

void seq::resize(int s)
{
  data.resize(s);
  // careful- these new cells are full or garbage!
}

void seq::join(const seq & rhs)
{
  int k;
  //  cout << "Pre joined: ";
  //  cout << *this << endl << rhs << endl;
  int oldSize = size();
  resize(size() + rhs.size());
  for (k=0; k<rhs.size(); k++)
    {
      data[ k + oldSize] = rhs[k];
    }
  //	cout << *this << endl << endl;
}

set seq::operator <(double d)
{
  // first count num greater than d...
  int k, count = 0;
  for (k=0; k<size(); k++)
    {
      if (data[k] < d)
	count++;
    }

  int c = 0;

  set ret(count);
  for (k=0; k<size(); k++)
    {
      if (data[k] < d)
	{
	  ret[c] = k;
	  c++;
	}
    }
  return ret;
}

set seq::operator >(double d)
{
  // first count num greater than d...
  int k, count = 0;
  for (k=0; k<size(); k++)
    {
      if (data[k] > d)
	count++;
    }

  int c = 0;

  set ret(count);
  for (k=0; k<size(); k++)
    {
      if (data[k] > d)
	{
	  ret[c] = k;
	  c++;
	}
    }
  return ret;
}

int seq::search(double d)
{
  int k;
  for (k=0; k<size(); k++)
    if (data[k] == d)
      return k;
  return -1;
}

void swap(seq & x, seq & y)
{
  // cell-wise swapping is faster than reallocation
	double temp;
	for (unsigned int k=0; k<x.size();  k++)
	{
	  swap(x[k], y[k]);
	}
}

ostream & operator << (ostream & os, seq rhs)
{
  int w = os.width();
  os << "[ ";
  for (int k=0; k<rhs.size(); k++)
    {
      os.width(5);
      os << '(' << k << ')' << ' ';
      os << rhs[k];
      if (k != rhs.size()-1)
	os << ", ";
    }
  os << " ]";
  os.width(w);
  return os;
}

double sum(const seq & rhs)
{
  int k;
  double tot = 0;
  for (k=0; k<rhs.size(); k++)
    {
      tot += rhs[k];
    }
  return tot;
}

double absSum(const seq & rhs)
{
  int k;
  double tot = 0;
  for (k=0; k<rhs.size(); k++)
    {
      tot += fabs(rhs[k]);
    }
  return tot;
}

double dotProd(const seq & lhs, const seq & rhs)
{
  int k;
  double tot = 0;
  for (k=0; k<rhs.size(); k++)
    {
      tot += lhs[k] * rhs[k];
    }
  return tot;
}

seq abs(const seq & rhs)
{
  seq temp = rhs;
  for (int k=0; k<temp.size(); k++)
    {
      temp[k] = fabs( temp[k] );
    }
  return temp;
}

double max(seq rhs)
{
  return rhs.max();
}

double min(seq rhs)
{
  return rhs.min();
}

double magDiffSquare(const seq & lhs, const seq & rhs)
{
  double tot = 0;
  for (int k=0; k<lhs.size(); k++)
    {
      tot += (lhs[k] - rhs[k]) * (lhs[k] - rhs[k]);
    }
  return tot;
  }

double magnitude(const seq & rhs)
{
  return sqrt(dotProd(rhs, rhs));
}
