/*******
svmvia, an implementation of the full regularization path for support vector machines
Copyright (C) 2007 Oliver Serang
Written by Oliver Serang at the Department of Genome Sciences at the University of Washington

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*********/

#include "svm.h"

SVM::SVM(const matrix & xVals, const seq & yVals, const matrix & vX, const seq & vY):
  kStar(1,1), kDiag(yVals.size())
{
  int size = yVals.size();

  cycle = 0;
  x = xVals;
  y = yVals;

  valX = vX;
  valY = vY;

  valN = valY.size();

  left = set(size);
  all = set(size);

  n = size;

  //  cout << "n = " << n << endl;
  max_cycles = n*3;

  nPlus = y.numMatch(1);
  nMinus = n - nPlus;

  nPlusVal = vY.numMatch(1);
  nMinusVal = valN - nPlusVal;

  //  cout << "SVM: n+ " << nPlus << "\tn-" << nMinus << endl;


  int k;
  for (k=0; k<n; k++)
    {
      kDiag[k] = kFunction(x[k], x[k]);
    }

  // space saver

  if (CACHE)
    {
  kernelMat = matrix(n, n);
  int j;
  for (k=0; k<n; k++)
    {
      for (j=0; j<n; j++)
	{
	  kernelMat[k][j] = kFunction(x[k], x[j]);
	}
    }
    }

  kStar[0][0] = 0;
}

double SVM::kernel(int k, int j)
{
  //  cout << "retrieving kerenel " << k << " " << j << endl;

  if (CACHE)
    return kernelMat[k][j];

  if (k == j)
    {
      return kDiag[k];
    }
  return kFunction( x[k], x[j]);
}

void SVM::computeInitialLargeF()
{
  int k,j;
  fLargeInit = seq(n);
  for (k=0; k<n; k++)
    {
      fLargeInit[k] = 0;
      for (j=0; j<n; j++)
	{
	  fLargeInit[k] += kernel(k, j) * y[j];
	}
    }
  //  return fLargeInit;
}

seq SVM::largeFInit(const set & s)
{
  // start with all alpha = 1 and alpha0 = 0
  seq f(s.size());
  int k, j;
  for (k=0; k<s.size(); k++)
    {
      f[k] = 0;
      for (j=0; j<s.size(); j++)
	{
	  f[k] += kernel(s[k], s[j]) * y[ s[j] ];
	}
    }
  return f;
}

seq SVM::getValF()
{
  seq f(valN);
  int k,j;
  for (k=0; k<valN; k++)
    {
      f[k] = 0;
      for (j=0; j<n; j++)
	{
	  if (alpha[j] != 0)
	    f[k] += kFunction(valX[k], x[j]) * y[j] * alpha[j];
	}
    }
  return (f + alpha0) / lambda;
}

seq SVM::getF()
{
  // used to set fL in the initialization

  seq f(n);
  int k,j;
  for (k=0; k<n; k++)
    {
      f[k] = 0;
      for (j=0; j<n; j++)
	{
	  if (alpha[j] != 0)
	    f[k] += kernel(k, j) * y[j] * alpha[j];
	}
    }
  return (f + alpha0) / lambda;
}

void SVM::balancedInit(const set & s)
{
  // space saver
  //  seq f = kernel[s] * y[s];

  //    cout << "Bal init..." << endl;

  // last working version
  //   seq f = largeFInit(s);
  seq f = fLargeInit[s];

  //  cout << "LARGE F:" << endl << f << endl;

  //  f = kernelMat[s] * y[s];

//  cout << "Ahem, real large f:" << endl;
//  cout << f << endl;

  set iPlus = y[s].match(1);
  set iMinus = y[s].match(-1);

  double fMax = f[iPlus].max();
  double fMin = f[iMinus].min();

//  cout << fMax << " " << fMin << endl;

  set maxS = iPlus[ f[iPlus].maxSet() ];
  set minS = iMinus[ f[iMinus].minSet() ];

  lambda = (fMax - fMin) / 2;
  double beta0 = 1-fMax/lambda;

  //  cout << "Setting alpha0 with this b0 = " << beta0 << endl;
  alpha0 = beta0 * lambda;

  elbow = maxS;
  elbow.join(minS);
}

void SVM::unbalancedInit()
{
  //    cerr << "Sorry, kiddo. This hasn't been done.\n\n\tYet..." << endl;
  //    cout << "Unbalanced initialization" << endl;
  cout << "Solving initial QP" << endl;
  SMO init(x, y);
  init.simpleSMO();
  cout << "Initial conditions set" << endl;

  alpha  = init.alpha;
  lambda = init.lambda;
  alpha0 = init.beta;

  //  cout << "SVM's Results from SMO: (lambda)" << lambda << endl;
  //  cout << "Results from SMO:" << lambda << endl;

  /***
  cout << "Unbal: " << endl;
  cout << "\t\t" << alpha0 << endl;
  cout << "alpha: " << endl;
  cout << alpha << endl;
  cout << "lambda " << lambda << endl;
  ***/

  init.createSets(left, elbow, right);
  removeLeft(elbow);
  removeLeft(right);
}

void SVM::addLeft(int i)
{
  for (int k=0; k<n; k++)
    {
      fLargeInit[k] += kernel(k, i) * y[i];
    }
}

void SVM::removeLeft(const set & s)
{
  for (int k=0; k<s.size(); k++)
    removeLeft(s[k]);
}

void SVM::removeLeft(int i)
{
  for (int k=0; k<n; k++)
    {
      fLargeInit[k] -= kernel(k, i) * y[i];
    }
}

void SVM::updateOptima()
{
  double err = error();
  //  cout << "Exe lambda, err = " << lambda << " " << err << endl;
  //    cout << "C: " << 1/lambda << " " << err << endl;
  //  cout << "Update: " << cycle << " " << lambda << endl;
  //  cout << cycle << '\t' << lambda << '\t' << err << endl;

  if ( VERBOSE )
    {
      cout << cycle << '\t' << 1/lambda << '\t' << err << endl;
    }

  /***
  cout << "Optima is comparing the cached validation pred. to the real" << endl;
  for (int k=0; k<valX.r; k++)
    {
      cout << k << "\t" << fValL[k] << " " << deNovo(valX[k]) << endl;
    }
  getchar();
  ***/

  //  cout << "SVM Validate: L, err = " << lambda << " " << err << endl;

  if (err < bestError)
    {
      alpha0Opt = alpha0;
      alphaOpt = alpha;
      lambdaOpt = lambda;
      
      bestError = err;
      //      cout << "New best error from L = " << lambdaOpt << endl;
    }
}

void SVM::fitPath()
{
  bestError = infinity;

  computeInitialLargeF();
  if (nPlus == nMinus)
    {
      balancedInit(all);

      removeLeft(elbow);
      left.remove(elbow);

      alpha = seq(n);
      alpha.setAll(1);
    }
  else
    {
      unbalancedInit();
    }
  lambdaOpt = lambda;
  alphaOpt = alpha;
  alpha0Opt = alpha0;

  //  cout << "Initial: " << lambda << endl;

  updateKStar(elbow);
  //  cout << "Init kStar: " << endl;
  //  cout << kStar << endl;
//  cout << "Baby's first K* is" << endl << kStar << endl;
  
  // space saver
//  fL = (kernel * (alpha*y) + alpha0) / lambda;

  fL = getF();
  fValL = getValF();

  //  cout << "FL INITIAL" << endl;
  //  cout << fL << endl << endl;
  
  cycle;
  for (cycle = 1; cycle <= max_cycles && lambda > LAMBDA_MIN; cycle++)
    {
      updateOptima();

      //	for (int RR = 0; RR<5; RR++)
      //		cout << endl;

      //      cout << "Starting cycle " << cycle << endl;

      //      cout << "lambda " << lambda << endl;
      //      cout << "alpha0 " << alpha0 << endl;
      //  cout << "Alpha\t" << alpha << endl;
      //  cout << "Alpha0\t" << alpha0 << endl;
      //cout << "lambda\t" << lambda << endl;
	  //cout << "left\t" << left << endl;
	  //	  cout << "elbow\t" << elbow << endl;
	  //cout << "right\t" << right << endl;
	  //	  cout << "K*" << endl << kStar << endl << endl;
	  //	  cout << "Fl: " << fL << endl;
	  //	  getchar();


	  //      	  cout << "fL\t" << fL << endl << endl;
	  
      if (elbow.size() == 0)
	{
		double alpha0Old = alpha0;
		double lambdaOld = lambda;
	  balancedInit(left);
	  // don't change alpha
	  elbow = left[elbow];
	  
	  removeLeft(elbow);
	  left.remove(elbow);

	  updateKStar(elbow);
	  fL = (fL+ ((alpha0 - alpha0Old)/lambdaOld) ) * lambdaOld / lambda;
	  fValL = (fValL + ((alpha0 - alpha0Old)/lambdaOld) ) * lambdaOld / lambda;
	}
      else
	{
	  double lambdaOld = lambda;


//		cout << "Solving K*" << endl;
//		cout << kStar << endl;

	  seq bStar = solveKStar();


	  /***
	  if (cycle > 15000)
	    {
	      cout << "Solving K*" << endl << endl;
	      cout << kStar << endl;
	      cout << "Solutiong = " << endl << bStar << endl;
	    }
	  ***/
	  
//	  cout << "B*\t" << bStar << endl;
	  
	  double b0 = bStar[0];
	  seq b(bStar.size()-1);
	  int k, j;
	  for (k=0; k<b.size(); k++)
	    {
	      b[k] = bStar[k+1];
	    }

	  // space saver
	  //	  gL = (kernel.restrictCol(elbow) * (y[elbow] * b)) + b0;
	  seq tempProd = y[elbow] * b;
	  gL.resize(n);
	  gValL.resize(valN);

	  for (k=0; k<valN; k++)
	    {
	      gValL[k] = 0;
	      for (j=0; j<elbow.size(); j++)
		{
		  gValL[k] += tempProd[j] * kFunction(valX[k], x[elbow[j]]);
		}
	    }
	  gValL += b0;

	  for (k=0; k<n; k++)
	    {
	      gL[k] = 0;
	      for (j=0; j<elbow.size(); j++)
		{
		  gL[k] += tempProd[j] * kernel(k, elbow[j]);
		}
	    }
	  gL += b0;

	  dL = fL - gL;
	  dValL = fValL - gValL;
	  bool immobile = (absSum(dL)/n < epsilon);
	  
//	  cout << "a, b*L" << endl;
//	  cout << alpha[elbow] << endl << b*lambda << endl;
	  seq temp = -alpha[elbow] + (b * lambda);
	  
//	  cout << "TEMP: " << temp << endl;
	  
	  seq lambdaLeft = (temp + 1) / b;
	  seq lambdaRight = temp / b;
	  
//	  cout << "Lambda LR" << endl;
//	  cout << lambdaLeft << endl << lambdaRight << endl << endl;
	  
	  for (k=0; k<temp.size(); k++)
	    {
	      if (fabs(b[k]) < epsilon)
		{
		  lambdaLeft[k] = -1;
		  lambdaRight[k] = -1;
		}
	    }

	  seq lambda01 = lambdaRight;
	  lambda01.join(lambdaLeft);
	  set cond = lambda01 < (lambda-epsilon);
//	  cout << "Cond " << cond << endl;
	  double lambdaExit = max(lambda01[cond]);

	  double lambdaEntry;
	  if (immobile && lambdaExit < epsilon)
	    {
	      cout << "Bailing out..." << endl;
	      break;
	    }
	  seq lambdaI;
	  if (!immobile)
	    {
	      lambdaI = dL / (y-gL) * lambda;
	      set tempSet = ( abs(y-gL) ) < epsilon;
	      for (k=0; k<tempSet.size(); k++)
		{
		  lambdaI[tempSet[k]] = -infinity;
		}

	      lambdaEntry = max(lambdaI[ lambdaI < (lambda - epsilon) ] );
	    }
	  else
	    lambdaEntry = -1;

	  double lambdaMax = max(lambdaEntry, lambdaExit);
	  lambda = lambdaMax;

	  for (k=0; k<elbow.size(); k++)
	    {
	      alpha[ elbow[k] ] = alpha[ elbow[k] ] - (lambdaOld - lambda) * b[k];
	    }
	  //	  cout << "Setting alpha0 with another b0 " << b0 << endl;
	  alpha0 = alpha0 - (lambdaOld - lambda) * b0;
	  fL = dL * lambdaOld / lambda + gL;
	  fValL = dValL * lambdaOld / lambda + gValL;

	  if (lambdaEntry > lambdaExit)
	    {
	      int i = lambdaI.search(lambdaEntry);
	      if ( left.contains(i))
		{
		  removeLeft(i);
		  left.remove(i);
		}
	      else
		{
		  right.remove(i);
		}
	      updateKStarOther(i);
	      elbow.add(i);
	    }
	  else
	    {
	      set i = elbow[ abs(lambdaRight - lambdaExit) < epsilon ];
	      set idrop;
	      seq leaveRight;
	      if (i.size() > 0)
		{
		  idrop = i;
		  leaveRight.resize(i.size());
		  leaveRight.setAll(1);
		}
	      i = elbow[ abs(lambdaLeft - lambdaExit) < epsilon ];
	      if (i.size() >0)
		{
		  idrop.join(i);
		  seq temp(i.size());
		  temp.setAll(0);
		  leaveRight.join(temp);
		}
	      set obs = idrop;
	      //cout << "Starting downdate process..." << endl;
	      //cout << "K* " << endl << kStar << endl;
	      //cout << "idrop = " << idrop << endl;

	      for (k=0; k<idrop.size(); k++)
		{
//			cout << "k is " << k << endl;
//			cout << "elbow is : " << elbow << endl;
		  if (leaveRight[k] == 1)
		    {
		      right.add(idrop[k]);
		    }
		  else
		    {
		      addLeft(idrop[k]);
		      left.add(idrop[k]);
		    }
		  int mi = elbow.search( idrop[k] );
		  
		  //cout << "mi is " << mi << endl;
		  downdateKStar(mi);
		  elbow.removeIndex(mi);
		}
	      //cout << "Finished with the downdate" << endl;
	      //cout << "K*" << endl;
	      //cout << kStar << endl << endl;
	    }
	}
    }
  cout << "Final cycle: " << cycle << endl;
  updateOptima();

  //	cout << "<fin>" << endl << "alpha" << endl << alpha << endl << endl << "alpha0" << endl << alpha0 << endl << endl;
  //	cout << "lambda" << endl << lambda << endl << endl;
}

      void SVM::downdateKStar(int i)
	{
	  // remove the row and column i
//	  cout << "K* pre downdate: " << i+1 << endl << kStar << endl;
	  kStar.removeRC(i+1);
//	  cout << "K* post downdate: " << i+1 << endl << kStar << endl;
	}

seq SVM::solveKStar()
{
  matrix kTemp = kStar;
  kTemp.invert();
  seq tempRes(kStar.c);
  tempRes.setAll(1);
  tempRes[0] = 0;
  seq b0 = kTemp * tempRes;
  return b0;
}

void SVM::updateKStar(const set & rhs)
{
  // basically create a new K*

  // check to see if your hypothesis about K* = {{0}} is correct...
  //  cout << "Remake K* with " << rhs << endl;
  if (kStar.r != 1 && kStar.c != 1)
    {
      cerr << "Serious problems here, over..." << endl;
    }

  // also assume that kStar[0][0] is the same...

  kStar.resize(rhs.size() + 1, rhs.size() + 1);
  int k,j;
  for (k=0; k<rhs.size(); k++)
    {
      kStar[k+1][0] = y[ rhs[k] ];
    }
  for (k=0; k<rhs.size(); k++)
    {
      kStar[0][k+1] = y[ rhs[k] ];
    }
  for (k=0; k<rhs.size(); k++)
    {
      for (j=0; j<rhs.size(); j++)
	{
	  kStar[k+1][j+1] = y[ rhs[k] ] * y[ rhs[j] ] * kernel( rhs[k] , rhs[j] );
	}
    }
}

void SVM::updateKStarOther(int i)
{
  // append K*
  //  matrix temp(kStar.r+1, Kstar.c+1);
  kStar.resize(kStar.r+1, kStar.c+1);
  int k;

  // it might be prudent to check the size of elbow here...

  //  cout << "Pre: " << endl;
  //  cout << kStar << endl;
  //  cout << "Elbow is: " << elbow << endl;

  for (k=0; k<elbow.size(); k++)
    {
      kStar[kStar.r-1][k+1] = kernel(i, elbow[k] ) * y[ elbow[k] ] * y[i];
    }
  for (k=0; k<elbow.size(); k++)
    {
      kStar[k+1][kStar.c-1] = kernel(i, elbow[k] ) * y[ elbow[k] ] * y[i];
    }
	kStar[0][kStar.c-1] = y[i];
	kStar[kStar.r-1][0] = y[i];
  kStar[kStar.r-1][kStar.c-1] = kernel(i, i) * y[i] * y[i];
}

double SVM::deNovo(const seq & xA)
{
  double tot = 0;
  for (int k=0; k<alpha.size(); k++)
    {
      if (alpha[k] != 0)
	tot += alpha[k] * kFunction(xA, x[k]) * y[k];
    }
  return (tot + alpha0) / lambda;
}

double SVM::error()
{
  if (ERR_MODE == 0)
    {
      return equalWeightError();
    }
  else if (ERR_MODE == 1)
    {
      return weightError();
    }
  else if (ERR_MODE == 2)
    {
      return rocError();
    }
  else
    {
      // invalid error mode...
      cerr << "Invalid validation mode specified" << endl << endl;
      exit(1);
    }
}

double SVM::weightError()
{
  double posErr = 0;
  double negErr = 0;
  for (int k=0; k<valX.r; k++)
    {
      if (sign(fValL[k]) != valY[k])
	{
	  if (valY[k] == 1)
	    {
	      posErr++;
	    }
	  else
	    {
	      negErr++;
	    }
	}
    }
  //  cout << "Weight error: " << ( (posErr * double(nMinus) / nPlus) + negErr ) / (2*nMinus) << endl;
  return ( (posErr * double(nMinusVal) / nPlusVal) + negErr ) / (2*nMinusVal);
}

double SVM::rocError()
{
  // could be sped up later by not keeping a local copy...
  vector<vector<double> > temp(valY.size());
  int k,j;
  for (k=0; k<valY.size(); k++)
    {
      temp[k] = vector<double> (2);

      // I think these should work lexicographically...
      temp[k][0] = fValL[k]*valY[k];
      temp[k][1] = valY[k];
    }

  sort( temp.begin(), temp.end() );
  /***
  for (k=0; k<valY.size(); k++)
    {
      for (j=0; j<2; j++)
	cout << temp[k][j] << " ";
      cout << endl;
    }
  cout << endl;
  getchar();
  ***/

  double area = 0;
  int cumPlus = 0;
  int cumMinus = 0;
  for (k=0; k<temp.size(); k++)
    {
      if (temp[k][1] == 1)
	{
	  cumPlus++;
	}
      else
	{
	  area += cumPlus;
	  cumMinus++;
	}
    }

  //  cout << "ROC err: " << 1 - area / (nPlus*nMinus) << endl;
  return 1 - area / (nPlusVal*nMinusVal);
}

double SVM::equalWeightError()
{
  double err = 0;
  for (int k=0; k<valX.r; k++)
    {
      if (sign(fValL[k]) != valY[k])
	{
	  err++;
	}
    }
  return err / valX.r;
}

double SVM::deNovoOpt(const seq & xA)
{
  double tot = 0;
  for (int k=0; k<alphaOpt.size(); k++)
    {
      if (alphaOpt[k] != 0)
	tot += alphaOpt[k] * kFunction(xA, x[k]) * y[k];
    }
  return (tot + alpha0Opt) / lambdaOpt;
}

double SVM::validate(const matrix & xU, const seq & yU)
{
  double err = 0;
  for (int k=0; k<xU.r; k++)
    {
      if (sign(deNovoOpt(xU[k])) != yU[k])
	{
	  err++;
	}
    }
  return err / xU.r;
}

